/*
 * Copyright (C) 2011 Canonical Ltd.
 *
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License version 3, as published
 * by the Free Software Foundation.

 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranties of
 * MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.

 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Authored by Ken VanDine <ken.vandine@canonical.com>
 */

using Dee;
using Gee;
using Config;
using Gwibber;
using Unity;

namespace Gwibber {

  /* DBus name for the place. Must match out .place file */
  const string BUS_NAME = "com.canonical.Unity.Gwibber";

  /**
   * The Daemon class implements all of the logic for the place.
   *
   */
  public class Daemon : GLib.Object
  {
    private Unity.PlaceController control;
    private Unity.PlaceEntryInfo place_entry;
    private Gwibber.Streams streams_service;
    private Dee.Model? model;

    construct
    {
      streams_service = new Gwibber.Streams();
      /* Create all the models we need to share with the Unity shell.
       * The model column schemas must align with what we defined in
       * schemas.vala */
      var sections_model = new Dee.SharedModel(BUS_NAME + ".SectionsModel");
      sections_model.set_schema ("s", "s");

      var groups_model = new Dee.SharedModel(BUS_NAME + ".GroupsModel");
      groups_model.set_schema ("s", "s", "s");

      var global_groups_model = new Dee.SharedModel(BUS_NAME + ".GlobalGroupsModel");
      global_groups_model.set_schema ("s", "s", "s");

      var results_model = new Dee.SharedModel(BUS_NAME + ".ResultsModel");
      results_model.set_schema ("s", "s", "u", "s", "s", "s");

      var global_results_model = new Dee.SharedModel(BUS_NAME + ".GlobalResultsModel");
      global_results_model.set_schema ("s", "s", "u", "s", "s", "s");
      
      /* Create a PlaceEntryInfo. The properties of the PlaceEntryInfo is
       * automatically mapped back and forth over the bus, and removes most
       * DBusisms from the place daemon code.
       *
       * Object path of the PlaceEntryInfo must match the one defined in the
       * the place entry in the place .place file
       */
      place_entry = new Unity.PlaceEntryInfo ("/com/canonical/unity/gwibber/main");
      place_entry.sections_model = sections_model;
      place_entry.entry_renderer_info.groups_model = groups_model;
      place_entry.entry_renderer_info.results_model = results_model;
      place_entry.global_renderer_info.groups_model = global_groups_model;
      place_entry.global_renderer_info.results_model = global_results_model;

      populate_sections ();
      populate_groups ();

      /* Unity preloads the icon defined in the .place file even before the
       * place daemon is running - when we come up it will ratify the icon */
      place_entry.icon = @"$(Config.PREFIX)/share/gwibber/unity/applications-microblogging.png";

      /* Listen for section changes and start updating our results model
       * accordingly */
      place_entry.notify["active-section"].connect (update_entry_results_model);

      /* Listen for changes to the place entry search */
      place_entry.notify["active-search"].connect (update_entry_results_model);

      /* Listen for changes to the global search */
      place_entry.notify["active-global-search"].connect (update_global_results_model);
      
      /* Listen for when the place is hidden/shown by the Unity Shell */
      place_entry.notify["active"].connect (
        (obj, pspec) => {
          debug (@"Activated: $(place_entry.active)");
          update_entry_results_model ();
        }
      );          

      /* The last thing we do is export the controller. Once that is up,
       * clients will expect the Dee.SharedModels to work.
       * The 'place' EntryInfo is exported for Unity to show by adding it to
       * the controller. You can add more than one EntryInfo here if you have
       * more than one place entry to export from the same place daemon */
      control = new Unity.PlaceController ("/com/canonical/unity/gwibber");
      control.add_entry (place_entry);
      
      /* Export the controller on the bus. Unity can see us past this point */
      try {
        control.export ();
      } catch (IOError error) {
        critical ("Failed to export DBus service for '%s': %s",
                  control.dbus_path, error.message);
      }

      var streams_model = streams_service.get_stream_model();
      model = streams_model;
    }

    private void populate_sections ()
    {
      var sections = place_entry.sections_model;

      if (sections.get_n_rows() != 0)
        {
          critical ("The sections model should be empty before initial population");
          sections.clear ();
        }

      /* The row offsets should match those from the Section enum */
      sections.append (_("Everything"), "");
      sections.append (_("Messages"), "");
      sections.append (_("Replies"), "");
      sections.append (_("Images"), "");
      sections.append (_("Videos"), "");
      sections.append (_("Links"), "");
      sections.append (_("Private"), "");
      sections.append (_("Public"), "");
      debug ("Sections has %u rows", sections.get_n_rows());
    }

    private void populate_groups ()
    {
      Icon icon;
      var groups = place_entry.entry_renderer_info.groups_model;

      if (groups.get_n_rows() != 0)
        {
          critical ("The groups model should be empty before initial population");
          groups.clear ();
        }

      /* The row offsets should match those from the Group enum */
      icon = new ThemedIcon ("tray-message");
      groups.append ("UnityHorizontalTileRenderer", _("Everything"), icon.to_string());
      icon = new ThemedIcon ("tray-message");
      groups.append ("UnityHorizontalTileRenderer", _("Messages"), icon.to_string());
      icon = new ThemedIcon ("mail_reply");
      groups.append ("UnityHorizontalTileRenderer", _("Replies"), icon.to_string());
      icon = new ThemedIcon ("gnome-mime-image");
      groups.append ("UnityHorizontalTileRenderer", _("Images"), icon.to_string());
      icon = new ThemedIcon ("video");
      groups.append ("UnityHorizontalTileRenderer", _("Videos"), icon.to_string());
      icon = new ThemedIcon ("web-browser");
      groups.append ("UnityHorizontalTileRenderer", _("Links"), icon.to_string());
      icon = new ThemedIcon ("mail-send-receive");
      groups.append ("UnityHorizontalTileRenderer", _("Private"), icon.to_string());
      icon = new ThemedIcon ("evolution");
      groups.append ("UnityHorizontalTileRenderer", _("Public"), icon.to_string());
      debug ("Groups has %u rows", groups.get_n_rows());
    }
    
    private void update_entry_results_model ()
    {
      /* If we're not active just ignore anything Unity tells us to do
       * it's not gonna show on the screen anyway */
      if (!place_entry.active)
        return;
                    
      var _results_model = place_entry.entry_renderer_info.results_model;
      var _groups_model = place_entry.entry_renderer_info.groups_model;
      var _search = place_entry.active_search;
      var _section = (Section)place_entry.active_section;          
       
      update_results_model (_results_model, _groups_model,
                            _search, _section);
    }
    
    private void update_global_results_model ()
    {
      /* Unlike update_entry_model() the global model may be updated
       * even though we are not active */
    
      var _results_model = place_entry.global_renderer_info.results_model;
      var _groups_model = place_entry.global_renderer_info.groups_model;
      var _search = place_entry.active_global_search;
      var _section = Section.ALL; /* We have no active section in global mode */
       
      update_results_model (_results_model, _groups_model,
                            _search, _section);
    }

    /* Generic method to update a results model. We do it like this to minimize
     * code dup between updating the global- and the entry results model */
    private void update_results_model (Dee.Model results_model, Dee.Model groups_model,
                                       Unity.PlaceSearch? search, Section section)
    {
      unowned Dee.ModelIter iter, end;
      string search_string = null;
      Group group = Group.ALL;
      debug ("Rebuilding results model");
      results_model.clear ();

      iter = model.get_first_iter ();
      end = model.get_last_iter ();
      
      place_entry.freeze_notify ();

      if (search != null)
      { 
        if (search.get_search_string () != null)
          search_string = search.get_search_string ().down();
      }

      while (iter != end)
      {
        if (section == Section.MESSAGES && model.get_string(iter, StreamModelColumn.STREAM) != "messages")
        {
          iter = model.next (iter);
          continue;
        }

        if (section == Section.REPLIES && model.get_string(iter, StreamModelColumn.STREAM) != "replies")
        {
          iter = model.next (iter);
          continue;
        }

        if (section == Section.IMAGES && model.get_string(iter, StreamModelColumn.STREAM) != "images")
        {
          iter = model.next (iter);
          continue;
        }

        if (section == Section.VIDEOS && model.get_string(iter, StreamModelColumn.STREAM) != "videos")
        {
          iter = model.next (iter);
          continue;
        }

        if (section == Section.LINKS && model.get_string(iter, StreamModelColumn.STREAM) != "links")
        {
          iter = model.next (iter);
          continue;
        }

        if (section == Section.PRIVATE && model.get_string(iter, StreamModelColumn.STREAM) != "private")
        {
          iter = model.next (iter);
          continue;
        }

        if (section == Section.PUBLIC && model.get_string(iter, StreamModelColumn.STREAM) != "public")
        {
          iter = model.next (iter);
          continue;
        }

        if (search_string == null || search_string in model.get_string(iter, StreamModelColumn.SENDER).down() || search_string in model.get_string(iter, StreamModelColumn.MESSAGE).down())
          {
          if (model.get_string(iter, StreamModelColumn.STREAM) == "messages")
            group = Group.MESSAGES;
          else if (model.get_string(iter, StreamModelColumn.STREAM) == "replies")
            group = Group.REPLIES;
          else if (model.get_string(iter, StreamModelColumn.STREAM) == "images")
            group = Group.IMAGES;
          else if (model.get_string(iter, StreamModelColumn.STREAM) == "videos")
            group = Group.VIDEOS;
          else if (model.get_string(iter, StreamModelColumn.STREAM) == "links")
            group = Group.LINKS;
          else if (model.get_string(iter, StreamModelColumn.STREAM) == "private")
            group = Group.PRIVATE;
          else if (model.get_string(iter, StreamModelColumn.STREAM) == "public")
            group = Group.PUBLIC;
          results_model.append(model.get_string(iter, StreamModelColumn.URL), model.get_string(iter, StreamModelColumn.ICON_URI), group, "text/html", model.get_string(iter, StreamModelColumn.SENDER), model.get_string(iter, StreamModelColumn.MESSAGE));
          }
        iter = model.next (iter);
      }
      if (iter == end)
      {
        iter = null;
      }
     
      Idle.add (() => {
        place_entry.thaw_notify ();
        return false;
      });

      if (search != null)
        search.finished ();

      debug ("Results has %u rows", results_model.get_n_rows());
    }

  } /* End Daemon class */
} /* end Gwibber namespace */
