/*
 * Copyright (C) 2010 Neil Jagdish Patel
 *
 * This library is free software; you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License
 * version 3.0 as published by the Free Software Foundation.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License version 3.0 for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library. If not, see
 * <http://www.gnu.org/licenses/>.
 *
 * Authored by Neil Jagdish Patel <njpatel@gmail.com>
 */

using Gwibber;

namespace GwibberGtk
{
  public class StreamView : Gtk.HBox
  {
    private Gtk.Adjustment     adjustment;
    private Gtk.VScrollbar     scrollbar;
    private Gtk.Layout         layout;
    private Gtk.VBox           view_box;

    private Dee.Model? _model = null;
    private Dee.Model? _stream_filter_model = null;
    private Dee.Filter *stream_filter;
    private Dee.Filter *sort_filter;
    private int _position = 0;
    private string _stream = "messages";
    private string _sort_order = "desc";
    private uint refresh_id = 0;
    private int tiles_visible = 0;

    public Dee.Model model {
      get { return _model; }
      set {
        if (_model is Dee.Model)
          {
            _model.row_added.disconnect (on_row_added);
            _model.row_removed.disconnect (on_row_removed);
            _model.row_changed.disconnect (on_row_changed);
            if (_model is Dee.SharedModel)
              _model.notify["synchronized"].disconnect (on_synchronized);
          }

        _model = value;
        _model.row_added.connect (on_row_added);
        _model.row_removed.connect (on_row_removed);
        _model.row_changed.connect (on_row_changed);
        if (_model is Dee.SharedModel)
          _model.notify["synchronized"].connect (on_synchronized);
      }
    }

    public Dee.Model stream_filter_model {
      get { return _stream_filter_model; }
      set {
        stream_filter = new Dee.Filter.for_key_column (StreamModelColumn.STREAM, stream);
        if (sort_order == "asc")
          sort_filter = new Dee.Filter.collator(StreamModelColumn.TIMESTAMP);
        else
          sort_filter = new Dee.Filter.collator_desc(StreamModelColumn.TIMESTAMP);
        var filtered_model = new Dee.FilterModel (stream_filter, model);
        _stream_filter_model = new Dee.FilterModel (sort_filter, filtered_model);
        
        refresh ();
        Idle.add (() => {
          adjustment.set_upper ((double)(_stream_filter_model.get_n_rows ()));
          return false;
          });
      }
    }

    public int position {
      get { return _position; }
      set {
        if (_position != value)
          {
            _position = value;
            _position = _position.clamp (0, (int)stream_filter_model.get_n_rows ());

            adjustment.set_value (_position);
            refresh ();
          }
      }
    }

    public string stream {
      get { return _stream; }
      set {
        if (_stream != value)
        {
          _stream = value;
          stream_filter_model = model;
        }
      }
    }


    public string sort_order {
      get { return _sort_order; }
      set {
        if (value != _sort_order)
        {
          _sort_order = value;
          stream_filter_model = model;
        }
      }
    }

    private SList<StreamViewTile> tiles;

    private int last_width = 0;
    private int last_height = 0;

    public StreamView ()
    {
      Object (homogeneous:false, spacing:0);
    }

    ~StreamView ()
    {
      _stream_filter_model = null;
      _model = null;
    }

    construct
    {
      adjustment = new Gtk.Adjustment (0, 0, 1, 1, 1, 1);
      adjustment.notify["value"].connect (()=> {
        position = (int)adjustment.get_value();
        refresh ();
      });

      layout = new Gtk.Layout (null, null);
      pack_start (layout, true, true, 0);
                 
      view_box = new Gtk.VBox (false, 0);
      layout.put (view_box, 0, 0);

      scrollbar = new Gtk.VScrollbar (adjustment);
      pack_start (scrollbar, false, false, 0);
     
      
      set_events(Gdk.EventMask.KEY_PRESS_MASK | Gdk.EventMask.SCROLL_MASK);
      scroll_event.connect ((source, event) => {
        Gdk.ScrollDirection direction = event.direction;
        if (event.direction == Gdk.ScrollDirection.UP)
        {
          if (position > 0)
          {
            adjustment.set_value(position - 1);
          }
        }
        if (event.direction == Gdk.ScrollDirection.DOWN)
        {
          if (position < (adjustment.upper - tiles_visible))
          {
            adjustment.set_value(position + 1);
          }
        }
      });

      key_press_event.connect ((source, event) => {
        var key = event.keyval;

        switch (event.keyval)
        {
          case 0xff55: // *_Page_Up
          case 0xff9a:
            position -= tiles_visible;
            break;
          case 0xff56: // *_Page_Down
          case 0xff9b:
            position += tiles_visible;
            break;
          case 0xff54: // *_Down
          case 0xff99:
            position += 1;
            break;
          case 0xff52: // *_Up
          case 0xff97:
            position -= 1;
            break;
          default:
            return false;
        }
        return true;
      });


      for (int i = 0; i < 10; i++)
        {
          var tile = new StreamViewTile ();
          tile.show_all ();
          view_box.pack_start (tile, false, false, 0);
          tiles.append (tile);
        }

      layout.size_allocate.connect_after ((a) => {
          if (last_width != a.width || last_height != a.height)
          {
            last_width = a.width;
            last_height = a.height;
            view_box.set_size_request (last_width, -1);

            refresh ();
          }
        });
    }

    /*
     * This is a wrapper method to ensure we don't get DOS'd by row-added signals
     */
    private void refresh ()
    {
      if (refresh_id == 0)
      {
        refresh_id = Idle.add (() => {
          do_refresh ();
          refresh_id = 0;
          return false;
        });
      }
    }

    /*
     * This refreshes the view with the latest model and position
     */
    private void do_refresh ()
    {
      int i = 0;

      fill_up_remaining_space ();
      if (!(stream_filter_model is Dee.FilterModel))
      {
        stream_filter_model = model;
      }
      uint n_rows = stream_filter_model.get_n_rows ();

      unowned Dee.ModelIter? iter;
      iter = stream_filter_model.get_iter_at_row (position);
      
      foreach (StreamViewTile tile in tiles)
      {
        if (iter != null && position + i < n_rows)
        {
          tile.set_details (stream_filter_model.get_string (iter, StreamModelColumn.ACCOUNT),
                            stream_filter_model.get_string (iter, StreamModelColumn.STREAM),
                            stream_filter_model.get_string (iter, StreamModelColumn.SERVICE),
                            stream_filter_model.get_string (iter, StreamModelColumn.SENDER),
                            stream_filter_model.get_bool (iter, StreamModelColumn.FROM_ME),
                            stream_filter_model.get_string (iter, StreamModelColumn.TIMESTAMP),
                            stream_filter_model.get_string (iter, StreamModelColumn.MESSAGE),
                            stream_filter_model.get_string (iter, StreamModelColumn.HTML),
                            stream_filter_model.get_string (iter, StreamModelColumn.ICON_URI),
                            stream_filter_model.get_string (iter, StreamModelColumn.URL),
                            stream_filter_model.get_string (iter, StreamModelColumn.SOURCE),
                            stream_filter_model.get_string (iter, StreamModelColumn.TIMESTRING),
                            stream_filter_model.get_string (iter, StreamModelColumn.REPLY_NICK),
                            stream_filter_model.get_string (iter, StreamModelColumn.REPLY_URL),
                            stream_filter_model.get_double (iter, StreamModelColumn.LIKES));
        }
        else
        {
            tile.reset ();
        }
        tiles_visible = i;
        i++;
        if (position + i > n_rows)
          iter = null;
        else
          iter = stream_filter_model.get_iter_at_row (position + i);
      }

      Idle.add (()=>{
        queue_draw ();
        return false;
      });
    }

    private void fill_up_remaining_space ()
    {
      Gtk.Allocation alloc;
      int overall_height = 0;
      bool one_was_hidden = false;

      layout.get_allocation (out alloc);

      foreach (StreamViewTile tile in tiles)
      {
        Gtk.Allocation a;

        if (tile.get_visible ())
        {
            tile.get_allocation (out a);
            overall_height += a.height;
        }
        else
        {
          one_was_hidden = true;
        }
      }

      if (alloc.height > overall_height && !one_was_hidden)
      {
          var tile = new StreamViewTile ();
          tile.show_all ();
          view_box.pack_start (tile, false, false, 0);
          tiles.append (tile);
          refresh();
      }
      while (alloc.height < overall_height)
      {
          Gtk.Allocation b;
          StreamViewTile tile = tiles.nth_data(tiles.length() - 1);
          tile.get_allocation (out b);
          overall_height -= b.height;
          tile.reset();
          tiles.remove(tile);
          tiles_visible = (int)tiles.length ();
      }

      Idle.add (()=>{
        queue_draw ();
        return false;
      });
    }

    private void on_row_added ()
    {
      refresh ();
    }

    private void on_row_removed ()
    {
      refresh ();
    }

    private void on_row_changed ()
    {
      refresh ();
    }

    private void on_synchronized ()
    {
      debug ("on_synchronized");
      stream_filter_model = model;
      adjustment.set_upper ((double)(stream_filter_model.get_n_rows ()));
      refresh ();
    }
  }
}
